import logging
from contextlib import closing
from pathlib import Path
from typing import Any, List

import duckdb
import mcp.server.stdio
import mcp.types as types
from mcp.server import Server
from duckdbmcp.lib.handler import DuckDbTools
from pydantic import AnyUrl

from duckdbmcp import Config

logger = logging.getLogger("duckdbmcp")
logger.info("Starting MCP DuckDB Server")


class DuckDBDatabase:
    def __init__(self, config: Config):
        self.config = config
        self.db_path = config.db_path
        self.db_in_memory = config.db_in_memory
        self.readonly = config.readonly

        if self.db_in_memory:
            self.handler = DuckDbTools(
                db_path=':memory:',
                read_only=self.readonly,
            )
        else:
            self.handler = DuckDbTools(
                db_path=self.db_path.as_uri(),
                read_only=self.readonly,
            )


async def main(config: Config):
    logger.info(f"Starting DuckDB MCP Server with DB path: {config.db_path}")

    db = DuckDBDatabase(config)
    server = Server("mcp-duckdb-server")

    logger.debug("Registering handlers")

    @server.list_resources()
    async def handle_list_resources() -> list[types.Resource]:
        """
        List available duckdb resources.
        """
        return []

    @server.read_resource()
    async def handle_read_resource(uri: AnyUrl) -> str:
        """
        Read a specific note's content by its URI.
        """
        return "No data"

    @server.list_prompts()
    async def handle_list_prompts() -> list[types.Prompt]:
        """
        List available prompts.
        """
        return []

    @server.get_prompt()
    async def handle_get_prompt(name: str, arguments: dict[str, str] | None) -> types.GetPromptResult:
        """
        Generate a prompt by combining arguments with server state.
        """

        return types.GetPromptResult(
            description="No",
            messages=[],
        )

    @server.list_tools()
    async def handle_list_tools() -> list[types.Tool]:
        """List available tools"""
        tools = [
            types.Tool(
                name="query",
                description="Execute a query on the DuckDB database",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "SQL query to execute",
                        },
                    },
                    "required": ["query"],
                },
            ),
            types.Tool(
                name="show_tables",
                description="Show all tables in the DuckDB database",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "include_views": {
                            "type": "boolean",
                            "description": "Whether to include views in the result",
                            "default": False
                        }
                    },
                    "required": []
                }
            ),
            types.Tool(
                name="describe_table",
                description="Describe a table in the DuckDB database",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "table": {
                            "type": "string",
                            "description": "Name of table to describe"
                        }
                    },
                    "required": ["table"]
                }
            ),
            types.Tool(
                name="inspect_query",
                description="Inspect a query in the DuckDB database",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "SQL query to inspect"
                        }
                    },
                    "required": ["query"]
                }
            ),
            types.Tool(
                name="create_table_from_path",
                description="Create a table from a file path",
                inputSchema={
                    "type": "object", 
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "Path to the file to load"
                        },
                        "table": {
                            "type": "string",
                            "description": "Optional table name to use"
                        },
                        "replace": {
                            "type": "boolean", 
                            "description": "Whether to replace existing table",
                            "default": False
                        }
                    },
                    "required": ["path"]
                }
            ),
            types.Tool(
                name="create_table_from_url",
                description="Create a table from a URL",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "url": {
                            "type": "string",
                            "description": "URL to the file to load"
                        },
                        "table": {
                            "type": "string",
                            "description": "Optional table name to use"
                        },
                        "replace": {
                            "type": "boolean",
                            "description": "Whether to replace existing table", 
                            "default": False
                        }
                    },
                    "required": ["url"]
                }
            ),
            types.Tool(
                name="create_table_from_s3",
                description="Create a table from an S3 path",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "S3 path to the file to load"
                        },
                        "table": {
                            "type": "string",
                            "description": "Optional table name to use"
                        }
                    },
                    "required": ["path"]
                }
            ),
            types.Tool(
                name="create_table_from_csv",
                description="Create a table from a CSV file",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "path": {
                            "type": "string",
                            "description": "Path to the CSV file"
                        },
                        "table": {
                            "type": "string",
                            "description": "Optional table name to use"
                        },
                        "delimiter": {
                            "type": "string",
                            "description": "Optional delimiter to use"
                        }
                    },
                    "required": ["path"]
                }
            ),
            types.Tool(
                name="summarize_table", 
                description="Get summary statistics for a table",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "table": {
                            "type": "string",
                            "description": "Name of table to summarize"
                        }
                    },
                    "required": ["table"]
                }
            ),
            types.Tool(
                name="export_table_to_path",
                description="Export a table to a file",
                inputSchema={
                    "type": "object",
                    "properties": {
                        "table": {
                            "type": "string", 
                            "description": "Name of table to export"
                        },
                        "format": {
                            "type": "string",
                            "description": "Format to export as (default: parquet)",
                            "default": "PARQUET"
                        },
                        "path": {
                            "type": "string",
                            "description": "Optional path to export to"
                        }
                    },
                    "required": ["table"]
                }
            )
        ]

        return tools

    @server.call_tool()
    async def handle_call_tool(
        name: str, arguments: dict[str, Any] | None
    ) -> list[types.TextContent | types.ImageContent | types.EmbeddedResource]:
        """Handle tool execution requests"""
        try:
            if not arguments:
                raise ValueError("Missing arguments")
            
            if name == "query":
                results = db.handler.run_query(arguments["query"])
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "show_tables":
                results = db.handler.show_tables(True)
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "describe_table":
                results = db.handler.describe_table(arguments["table"])
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "inspect_query":
                results = db.handler.inspect_query(arguments["query"])
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "create_table_from_path":  
                results = db.handler.create_table_from_path(
                    path=arguments["path"],
                    table=arguments.get("table"),
                    replace=arguments.get("replace", False)
                )
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "create_table_from_url":
                results = db.handler.create_table_from_path(
                    path=arguments["url"],
                    table=arguments.get("table"),
                    replace=arguments.get("replace", False)
                )
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "create_table_from_s3":
                results = db.handler.load_s3_path_to_table(
                    path=arguments["path"],
                    table=arguments.get("table")
                )
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "create_table_from_csv":
                results = db.handler.load_local_csv_to_table(
                    path=arguments["path"],
                    table=arguments.get("table"),
                    delimiter=arguments.get("delimiter")
                )
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "summarize_table":
                results = db.handler.summarize_table(arguments["table"])
                return [types.TextContent(type="text", text=str(results))]
            
            elif name == "export_table_to_path":
                results = db.handler.export_table_to_path(
                    table=arguments["table"],
                    format=arguments.get("format", "PARQUET"),
                    path=arguments.get("path")
                )
                return [types.TextContent(type="text", text=str(results))]
            else:
                raise ValueError(f"Unknown tool: {name}")
        except duckdb.Error as e:
            return [types.TextContent(type="text", text=f"Database error: {str(e)}")]
        except Exception as e:
            return [types.TextContent(type="text", text=f"Error: {str(e)}")]

    # Run the server using stdin/stdout streams
    options = server.create_initialization_options()
    async with mcp.server.stdio.stdio_server() as (read_stream, write_stream):
        logger.info("DuckDB MCP Server running with stdio transport")
        await server.run(
            read_stream,
            write_stream,
            options,
        )



"""
Test suite for the DuckDB-based MCP server module.
Uses pytest for running the tests.
"""

import tempfile
from pathlib import Path

import duckdb
import pytest

from duckdbmcp.config import Config
from duckdbmcp.server import DuckDBDatabase


def test_create_database_in_nonexistent_directory():
    """
    Ensure that when running in non-readonly mode, a missing directory
    can be created, and the database file is also created.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        db_path = Path(tmpdir) / "subdir" / "test.db"
        config = Config(db_path=db_path, readonly=False, db_in_memory=False)
        db = DuckDBDatabase(config)
        assert db_path.exists(), "Database file should be created automatically in non-readonly mode."
        # Optionally verify basic queries:
        db.handler.run_query("CREATE TABLE test (id INTEGER)")
        db.handler.run_query("INSERT INTO test VALUES (1)")
        result = db.handler.run_query("SELECT * FROM test")
        assert result == [(1,)], "Should be able to insert data when readonly=False."


def test_readonly_mode_missing_directory():
    """
    If the target directory does not exist and readonly=True,
    an exception should be raised.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        db_path = Path(tmpdir) / "subdir" / "test.db"
        config = Config(db_path=db_path, readonly=True, db_in_memory=False)
        # The directory 'subdir' does not exist, so it should raise an error.
        with pytest.raises(ValueError, match="Database directory does not exist"):
            DuckDBDatabase(config)


def test_readonly_mode_missing_db_file():
    """
    If the directory exists but the DB file itself does not exist and readonly=True,
    an exception should be raised.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        subdir = Path(tmpdir) / "subdir"
        subdir.mkdir(exist_ok=True)
        db_path = subdir / "test.db"
        config = Config(db_path=db_path, readonly=True, db_in_memory=False)
        with pytest.raises(ValueError, match="Database file does not exist"):
            DuckDBDatabase(config)


def test_normal_mode_create_and_write():
    """
    In non-readonly mode, we should be able to create a table and perform write operations.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        db_path = Path(tmpdir) / "test.db"
        config = Config(db_path=db_path, readonly=False, db_in_memory=False)
        db = DuckDBDatabase(config)

        db.handler.run_query("CREATE TABLE users (id INTEGER, name VARCHAR)")
        db.handler.run_query("INSERT INTO users VALUES (1, 'Alice'), (2, 'Bob')")

        result = db.handler.run_query("SELECT * FROM users ORDER BY id")
        assert result == [(1, "Alice"), (2, "Bob")], "Data insertion in non-readonly mode should succeed."


def test_readonly_mode_write_query_error():
    """
    In readonly mode, attempting a write operation (INSERT) should raise a duckdb.Error.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        # First, create the database in writable mode
        db_path = Path(tmpdir) / "test.db"
        config_writable = Config(db_path=db_path, readonly=False, db_in_memory=False)
        db_writable = DuckDBDatabase(config_writable)
        db_writable.handler.run_query("CREATE TABLE test (val INTEGER)")

        # Now open the same DB in readonly mode
        config_readonly = Config(db_path=db_path, readonly=True, db_in_memory=False)
        db_readonly = DuckDBDatabase(config_readonly)

        # Attempting to write should fail
        with pytest.raises(duckdb.Error):
            db_readonly.handler.run_query("INSERT INTO test VALUES (42)")


def test_readonly_flag_behavior():
    """
    Tests behavior for readonly=True vs. readonly=False.

    1) readonly=False: should allow table creation and data insertion.
    2) readonly=True: should raise an error when attempting to insert data.
                      use SELECT/SHOW/PRAGMA to verify read access.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        # 1) Writable configuration
        db_path_writable = Path(tmpdir) / "writable.db"
        config_writable = Config(db_path=db_path_writable, readonly=False, db_in_memory=False)
        db_writable = DuckDBDatabase(config_writable)

        db_writable.handler.run_query("CREATE TABLE items (id INTEGER)")
        db_writable.handler.run_query("INSERT INTO items VALUES (100)")
        result = db_writable.handler.run_query("SELECT * FROM items")
        assert result == [(100,)], "Should successfully write when readonly=False"

        # 2) Readonly configuration
        db_path_readonly = Path(tmpdir) / "readonly.db"
        # Create the DB file in non-readonly mode first
        config_init = Config(db_path=db_path_readonly, readonly=False, db_in_memory=False)
        db = DuckDBDatabase(config_init)
        db.handler.run_query("CREATE TABLE ro_test (val INTEGER)")
        db.handler.run_query("INSERT INTO ro_test VALUES (100)")

        # Reopen in readonly mode
        config_ro = Config(db_path=db_path_readonly, readonly=True, db_in_memory=False)
        db_ro = DuckDBDatabase(config_ro)

        # Attempt to write should fail
        with pytest.raises(duckdb.Error):
            db_ro.handler.run_query("INSERT INTO ro_test VALUES (999)")

        # Check that we can still read
        result = db_ro.handler.run_query("SELECT * FROM ro_test;")
        assert result == [(100,)]

        result = db_ro.handler.run_query("PRAGMA table_info(ro_test);")
        assert result == [(0, "val", "INTEGER", False, None, False)]

        result = db_ro.handler.run_query("SHOW TABLES;")
        assert result == [("ro_test",)]



import sys
from unittest.mock import patch

import pytest

from duckdbmcp.config import Config


def test_config_required_db_path():
    with patch.object(sys, "argv", ["prog"]):
        with pytest.raises(SystemExit):
            Config.from_arguments()


def test_config_from_arguments_readonly_false(tmp_path):
    test_db_file = tmp_path / "test.duckdb"
    test_args = [
        "--db-path",
        str(test_db_file),
    ]
    with patch.object(sys, "argv", ["prog"] + test_args):
        config = Config.from_arguments()
        assert config.db_path == test_db_file
        assert config.readonly is False


def test_config_from_arguments_readonly_true(tmp_path):
    test_db_file = tmp_path / "test.duckdb"
    test_args = [
        "--db-path",
        str(test_db_file),
        "--readonly",
    ]
    with patch.object(sys, "argv", ["prog"] + test_args):
        config = Config.from_arguments()
        assert config.readonly is True



import asyncio

from duckdbmcp.config import Config

from . import server


def main():
    """Main entry point for the package."""

    config = Config.from_arguments()
    asyncio.run(server.main(config))


# Optionally expose other important items at package level
__all__ = ["main", "server"]
